from __future__ import annotations

import argparse
import json
from typing import List

from together.files import Files


def add_parser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    COMMAND_NAME = "files"
    parser = subparsers.add_parser(COMMAND_NAME)

    child_parsers = parser.add_subparsers(required=True)

    _add_list(child_parsers, parents=parents)
    _add_upload(child_parsers, parents=parents)
    _add_delete(child_parsers, parents=parents)
    _add_retrieve(child_parsers, parents=parents)
    _add_retrieve_content(child_parsers, parents=parents)


def _add_list(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    list_parser = parser.add_parser("list", parents=parents)
    list_parser.set_defaults(func=_run_list)


def _add_upload(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    upload_file_parser = parser.add_parser("upload", parents=parents)
    upload_file_parser.add_argument(
        "file",
        metavar="FILENAME",
        help="Local file to upload",
        type=str,
    )
    upload_file_parser.set_defaults(func=_run_upload)


def _add_delete(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    delete_file_parser = parser.add_parser("delete", parents=parents)
    delete_file_parser.add_argument(
        "file_id",
        metavar="FILE-ID",
        help="File ID of remote file",
        type=str,
    )
    delete_file_parser.set_defaults(func=_run_delete)


def _add_retrieve(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    retrieve_file_parser = parser.add_parser("retrieve", parents=parents)
    retrieve_file_parser.add_argument(
        "file_id",
        metavar="FILE-ID",
        help="File ID of remote file",
        type=str,
    )
    retrieve_file_parser.set_defaults(func=_run_retrieve)


def _add_retrieve_content(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    retrieve_file_content_parser = parser.add_parser(
        "retrieve-content", parents=parents
    )
    retrieve_file_content_parser.add_argument(
        "file_id",
        metavar="FILE-ID",
        help="File ID of remote file",
        type=str,
    )
    retrieve_file_content_parser.add_argument(
        "--output",
        "-o",
        default=None,
        metavar="OUT_FILENAME",
        help="Optional output filename. Defaults to remote filename.",
        type=str,
        required=False,
    )

    retrieve_file_content_parser.set_defaults(func=_run_retrieve_content)


def _run_list(args: argparse.Namespace) -> None:
    files = Files(args.endpoint, log_level=args.log)
    response = files.list_files()
    print(json.dumps(response, indent=4))


def _run_upload(args: argparse.Namespace) -> None:
    files = Files(args.endpoint, log_level=args.log)
    response = files.upload_file(args.file)
    print(json.dumps(response, indent=4))


def _run_delete(args: argparse.Namespace) -> None:
    files = Files(args.endpoint, log_level=args.log)
    response = files.delete_file(args.file_id)
    print(json.dumps(response, indent=4))


def _run_retrieve(args: argparse.Namespace) -> None:
    files = Files(args.endpoint, log_level=args.log)
    response = files.retrieve_file(args.file_id)
    print(json.dumps(response, indent=4))


def _run_retrieve_content(args: argparse.Namespace) -> None:
    files = Files(args.endpoint, log_level=args.log)
    output = files.retrieve_file_content(args.file_id, args.output)
    print(output)



#! python
import argparse

from together.commands import api, files, finetune, inference
from together.utils import exit_1, get_logger


def main() -> None:
    parser = argparse.ArgumentParser(
        description="CLI client for Together API",
        prog="together",
    )

    parser.add_argument(
        "--endpoint",
        "-e",
        help="[Optional] Together API Endpoint URL",
        type=str,
        required=False,
        default=None,
    )

    base_subparser = argparse.ArgumentParser(add_help=False)
    base_subparser.add_argument(
        "--log",
        default="WARNING",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
        type=str,
        help="Set logging level. Defaults to WARNING. DEBUG will show all logs.",
        required=False,
    )

    subparser = parser.add_subparsers(dest="base")

    api.add_parser(subparser, parents=[base_subparser])
    inference.add_parser(subparser, parents=[base_subparser])
    finetune.add_parser(subparser, parents=[base_subparser])
    files.add_parser(subparser, parents=[base_subparser])

    args = parser.parse_args()

    # Setup logging
    try:
        logger = get_logger(__name__, log_level=args.log)
    except Exception:
        logger = get_logger(__name__, log_level="WARNING")

    try:
        args.func(args)
    except AttributeError as e:
        # print error, but ignore if `together` is run.
        if str(e) != "'Namespace' object has no attribute 'func'":
            logger.critical(f"Error raised: {e}")
            exit_1(logger)
        parser.print_help()


if __name__ == "__main__":
    main()



import logging
import sys
from typing import Optional


# Setup logging
def get_logger(
    name: str,
    logger: Optional[logging.Logger] = None,
    log_level: str = "WARNING",
) -> logging.Logger:
    if logger is None:
        logger = logging.getLogger(name)
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            handlers=[logging.StreamHandler(sys.stdout)],
        )

        logger.setLevel(log_level)

    return logger


def exit_1(logger: logging.Logger) -> None:
    logger.critical("Exiting with code 1...")
    sys.exit(1)



from __future__ import annotations

import argparse
import json
from typing import List

from together.api import API


def add_parser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    COMMAND_NAME = "api"
    parser = subparsers.add_parser(COMMAND_NAME)

    child_parsers = parser.add_subparsers(required=True)

    _add_list(child_parsers, parents=parents)
    _add_raw(child_parsers, parents=parents)


def _add_list(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    list_model_subparser = parser.add_parser("list", parents=parents)
    list_model_subparser.add_argument(
        "--all",
        "-a",
        help="List all models (available and unavailable)",
        default=False,
        action="store_true",
    )
    list_model_subparser.set_defaults(func=_run_list)


def _add_raw(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    list_parser = parser.add_parser("raw-supply", parents=parents)
    list_parser.set_defaults(func=_run_raw)


def _run_list(args: argparse.Namespace) -> None:
    api = API(endpoint_url=args.endpoint, log_level=args.log)

    if args.all:
        response = api.get_all_models()
    else:
        response = api.get_available_models()

    print(json.dumps(response, indent=4))


def _run_raw(args: argparse.Namespace) -> None:
    api = API(endpoint_url=args.endpoint, log_level=args.log)
    response = api.get_supply()
    print(json.dumps(response, indent=4))



from __future__ import annotations

import argparse
import base64
import json
import re
import sys
from typing import List

from together.inference import Inference
from together.utils import exit_1, get_logger


def add_parser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    COMMAND_NAME = "complete"
    inf_parser = subparsers.add_parser(COMMAND_NAME, parents=parents)

    inf_parser.add_argument(
        "prompt",
        metavar="PROMPT",
        default=None,
        type=str,
        help="A string providing context for the model to complete.",
    )

    inf_parser.add_argument(
        "--model",
        "-m",
        default="togethercomputer/RedPajama-INCITE-7B-Chat",
        type=str,
        help="The name of the model to query. Default='togethercomputer/RedPajama-INCITE-7B-Chat'",
    )

    inf_parser.add_argument(
        "--task",
        default="text2text",
        type=str,
        help="Task type: text2text, text2img. Default=text2text",
        choices=["text2text", "text2img"],
    )

    text2textargs = inf_parser.add_argument_group("Text2Text Arguments")

    text2textargs.add_argument(
        "--max-tokens",
        default=128,
        type=int,
        help="Maximum number of tokens to generate. Default=128",
    )
    text2textargs.add_argument(
        "--stop",
        default=["<human>"],
        nargs="+",
        type=str,
        help="Strings that will truncate (stop) inference text output. Default='<human>'",
    )
    text2textargs.add_argument(
        "--temperature",
        default=0.7,
        type=float,
        help="Determines the degree of randomness in the response. Default=0.7",
    )
    text2textargs.add_argument(
        "--top-p",
        default=0.7,
        type=float,
        help="Used to dynamically adjust the number of choices for each predicted token based on the cumulative probabilities. Default=0.7",
    )
    text2textargs.add_argument(
        "--top-k",
        default=50,
        type=int,
        help="Used to limit the number of choices for the next predicted word or token. Default=50",
    )
    text2textargs.add_argument(
        "--repetition-penalty",
        default=None,
        type=float,
        help="Controls the diversity of generated text by reducing the likelihood of repeated sequences. Higher values decrease repetition.",
    )
    text2textargs.add_argument(
        "--logprobs",
        default=None,
        type=int,
        help="Specifies how many top token log probabilities are included in the response for each token generation step.",
    )
    text2textargs.add_argument(
        "--raw",
        default=False,
        action="store_true",
        help="temperature for the LM",
    )

    text2imgargs = inf_parser.add_argument_group("Text2Image Arguments")

    text2imgargs.add_argument(
        "--height",
        default=512,
        type=int,
        help="Pixel height for generated image results",
    )
    text2imgargs.add_argument(
        "--width",
        default=512,
        type=int,
        help="Pixel width for generated image results",
    )
    # arguments for text2img models
    text2imgargs.add_argument(
        "--steps",
        default=50,
        type=int,
        help="Number of steps",
    )
    text2imgargs.add_argument(
        "--seed",
        default=42,
        type=int,
        help="Seed for image generation",
    )
    text2imgargs.add_argument(
        "--results",
        "-r",
        default=1,
        type=int,
        help="Number of image results to return",
    )
    text2imgargs.add_argument(
        "--output-prefix",
        "-o",
        default="image",
        type=str,
        help="Prefix for the file names the output images will be saved to. An image number will be appended to this name. Default=image",
    )

    inf_parser.set_defaults(func=_run_complete)


def _enforce_stop_tokens(text: str, stop: List[str]) -> str:
    """Cut off the text as soon as any stop words occur."""
    return re.split("|".join(stop), text)[0]


def _run_complete(args: argparse.Namespace) -> None:
    logger = get_logger(__name__, log_level=args.log)

    inference = Inference(
        endpoint_url=args.endpoint,
        task=args.task,
        model=args.model,
        max_tokens=args.max_tokens,
        stop=args.stop,
        temperature=args.temperature,
        top_p=args.top_p,
        top_k=args.top_k,
        repetition_penalty=args.repetition_penalty,
        logprobs=args.logprobs,
        steps=args.steps,
        seed=args.seed,
        results=args.results,
        height=args.height,
        width=args.width,
    )

    response = inference.inference(prompt=args.prompt)

    if args.raw:
        print(json.dumps(response, indent=4))
        sys.exit()

    if "output" in response.keys():
        if args.task == "text2text":
            try:
                text = str(response["output"]["choices"][0]["text"])
            except Exception:
                try:
                    logger.critical(f"Error raised: {response['output']['error']}")
                    exit_1(logger)
                except Exception as e:
                    logger.critical(f"Error raised: {e}")
                    exit_1(logger)

            # if args.stop is not None:
            #    text = _enforce_stop_tokens(text, args.stop)

        elif args.task == "text2img":
            try:
                images = response["output"]["choices"]

                for i in range(len(images)):
                    with open(f"{args.output}-{i}.png", "wb") as f:
                        f.write(base64.b64decode(images[i]["image_base64"]))
            except Exception as e:  # This is the correct syntax
                logger.critical(f"Error raised: {e}")
                exit_1(logger)

            text = f"Output images saved to {args.output}-X.png"
        else:
            logger.critical(
                f"Invalid task: {args.task}. Pick from either text2text or text2img."
            )
            exit_1(logger)

    elif "error" in response.keys():
        if response["error"] == "Returned error: no instance":
            logger.critical(
                f"No running instances for {args.model}. You can start an instance by navigating to the Together Playground at api.together.xyz"
            )
            exit_1(logger)
        else:
            logger.critical(f"Error raised: {response['error']}")

    else:
        logger.critical("Unknown response received")
        exit_1(logger)

    print(text.strip())



import os
import urllib.parse
from typing import Any, Dict, Optional

import requests

from together.utils import exit_1, get_logger


DEFAULT_ENDPOINT = "https://api.together.xyz/"


class Inference:
    def __init__(
        self,
        endpoint_url: Optional[str] = None,
        log_level: str = "WARNING",
        task: Optional[str] = "text2text",
        model: Optional[str] = None,
        max_tokens: Optional[int] = 128,
        stop: Optional[str] = None,
        temperature: Optional[float] = 0.7,
        top_p: Optional[float] = 0.7,
        top_k: Optional[int] = 50,
        repetition_penalty: Optional[float] = None,
        logprobs: Optional[int] = None,
        raw: Optional[bool] = False,
        # TODO stream_tokens: Optional[bool] = None
        steps: Optional[int] = 50,
        seed: Optional[int] = 42,
        results: Optional[int] = 1,
        height: Optional[int] = 512,
        width: Optional[int] = 512,
    ) -> None:
        # Setup logger
        self.logger = get_logger(str(__name__), log_level=log_level)

        together_api_key = os.environ.get("TOGETHER_API_KEY", None)
        if together_api_key is None:
            self.logger.critical(
                "TOGETHER_API_KEY not found. Please set it as an environment variable."
            )
            exit_1(self.logger)

        if endpoint_url is None:
            endpoint_url = DEFAULT_ENDPOINT

        self.together_api_key = together_api_key
        self.endpoint_url = urllib.parse.urljoin(endpoint_url, "/api/inference")

        self.task = task
        self.model = model
        self.max_tokens = max_tokens
        self.stop = stop
        self.temperature = temperature
        self.top_p = top_p
        self.top_k = top_k
        self.repetition_penalty = repetition_penalty
        self.logprobs = logprobs

        # text2img arguments
        self.steps = steps
        self.seed = seed
        self.results = results
        self.height = height
        self.width = width

    def inference(
        self,
        prompt: str,
    ) -> Dict[str, Any]:
        if self.task == "text2text":
            parameter_payload = {
                "model": self.model,
                "prompt": prompt,
                "top_p": self.top_p,
                "top_k": self.top_k,
                "temperature": self.temperature,
                "max_tokens": self.max_tokens,
                "stop": self.stop,
                "repetition_penalty": self.repetition_penalty,
                "logprobs": self.logprobs,
            }
        elif self.task == "text2img":
            parameter_payload = {
                "model": self.model,
                "prompt": prompt,
                "n": self.results,
                "mode": self.task,
                "steps": self.steps,
                "seed": self.seed,
                "height": self.height,
                "width": self.width,
            }
        else:
            self.logger.critical(
                f"Invalid task: {self.task}. Pick from either text2text or text2img."
            )
            exit_1(self.logger)

        # HTTP headers for authorization
        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
            "Content-Type": "application/json",
        }

        # send request
        try:
            response = requests.post(
                self.endpoint_url, headers=headers, json=parameter_payload
            )
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json



import json
import os
import posixpath
import urllib.parse
from logging import Logger
from typing import Dict, List, Optional, Union

import requests
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from together.utils import exit_1, get_logger


DEFAULT_ENDPOINT = "https://api.together.xyz/"


class JSONException(Exception):
    pass


def validate_file(file: str, logger: Logger) -> bool:
    if not os.path.isfile(file):
        logger.critical("ERROR: File not found")
        return False

    file_size = os.stat(file).st_size

    if file_size > 4.9 * (2**30):
        logger.warning("File size > 4.9 GB, file may fail to upload.")

    with open(file) as f:
        try:
            for line in f:
                json_line = json.loads(line)
                if "text" not in json_line:
                    logger.critical(
                        "ERROR: 'text' field not found in one or more lines in JSONL file"
                    )
                    return False
        except ValueError:
            logger.critical("ERROR: Could not load JSONL file. Invalid format")
            return False
        return True


class Files:
    def __init__(
        self,
        endpoint_url: Optional[str] = None,
        log_level: str = "WARNING",
    ) -> None:
        # Setup logger
        self.logger = get_logger(__name__, log_level=log_level)

        self.together_api_key = os.environ.get("TOGETHER_API_KEY", None)
        if self.together_api_key is None:
            self.logger.critical(
                "TOGETHER_API_KEY not found. Please set it as an environment variable."
            )
            exit_1(self.logger)

        if endpoint_url is None:
            endpoint_url = DEFAULT_ENDPOINT

        self.endpoint_url = urllib.parse.urljoin(endpoint_url, "/v1/files/")

    def list_files(self) -> Dict[str, List[Dict[str, Union[str, int]]]]:
        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            response = requests.get(self.endpoint_url, headers=headers)
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)
        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def upload_file(self, file: str) -> Dict[str, Union[str, int]]:
        try:
            data = {"purpose": "fine-tune", "file_name": os.path.basename(file)}

            headers = {
                "Authorization": f"Bearer {self.together_api_key}",
            }

            if not validate_file(file=file, logger=self.logger):
                exit_1(self.logger)

            session = requests.Session()

            init_endpoint = self.endpoint_url[:-1]

            self.logger.debug(
                f"Upload file POST request: data={data}, headers={headers}, URL={init_endpoint}, allow_redirects=False"
            )

            response = session.post(
                init_endpoint,
                data=data,
                headers=headers,
                allow_redirects=False,
            )

            self.logger.debug(f"Response text: {response.text}")
            self.logger.debug(f"Response header: {response.headers}")
            self.logger.debug(f"Response status code: {response.status_code}")

            if response.status_code == 401:
                self.logger.critical(
                    "This job would exceed your free trial credits. Please upgrade to a paid account through Settings -> Billing on api.together.ai to continue."
                )
                exit_1(self.logger)
            elif response.status_code != 302:
                self.logger.critical(
                    f"Unexpected error raised by endpoint. Response status code: {response.status_code}"
                )
                exit_1(self.logger)

            r2_signed_url = response.headers["Location"]
            file_id = response.headers["X-Together-File-Id"]

            self.logger.info(f"R2 Signed URL: {r2_signed_url}")
            self.logger.info("File-ID")

            self.logger.info("Uploading file...")

            file_size = os.stat(file).st_size
            with open(file, "rb") as f:
                with tqdm(
                    total=file_size, unit="B", unit_scale=True, unit_divisor=1024
                ) as t:
                    wrapped_file = CallbackIOWrapper(t.update, f, "read")
                    response = requests.put(r2_signed_url, data=wrapped_file)

            self.logger.info("File uploaded.")
            self.logger.debug(f"status code: {response.status_code}")
            self.logger.info("Processing file...")
            preprocess_url = urllib.parse.urljoin(
                self.endpoint_url, f"{file_id}/preprocess"
            )

            response = session.post(
                preprocess_url,
                headers=headers,
            )

            self.logger.info("File processed")
            self.logger.debug(f"Status code: {response.status_code}")

        except Exception as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        return {
            "filename": os.path.basename(file),
            "id": str(file_id),
            "object": "file",
        }

    def delete_file(self, file_id: str) -> Dict[str, str]:
        delete_url = urllib.parse.urljoin(self.endpoint_url, file_id)

        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            response = requests.delete(delete_url, headers=headers)
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def retrieve_file(self, file_id: str) -> Dict[str, Union[str, int]]:
        retrieve_url = urllib.parse.urljoin(self.endpoint_url, file_id)

        self.logger.info(f"Retrieve URL: {retrieve_url}")

        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            response = requests.get(retrieve_url, headers=headers)
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def retrieve_file_content(
        self, file_id: str, output: Union[str, None] = None
    ) -> str:
        if output is None:
            output = file_id + ".jsonl"

        relative_path = posixpath.join(file_id, "content")
        retrieve_url = urllib.parse.urljoin(self.endpoint_url, relative_path)

        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            session = requests.Session()

            response = session.get(retrieve_url, headers=headers, stream=True)
            response.raise_for_status()

            total_size_in_bytes = int(response.headers.get("content-length", 0))
            block_size = 1024 * 1024  # 1 MB
            progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)

            with open(output, "wb") as file:
                for chunk in response.iter_content(block_size):
                    progress_bar.update(len(chunk))
                    file.write(chunk)

            progress_bar.close()

            if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
                self.logger.warning(
                    "Caution: Downloaded file size does not match remote file size."
                )

        except requests.exceptions.RequestException as e:  # This is the correct syntax
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        return output  # this should be null



import os
from typing import Any, Dict, List, Optional, cast

import requests

from together.files import Files
from together.finetune import Finetune
from together.inference import Inference
from together.utils import exit_1, get_logger


DEFAULT_ENDPOINT = "https://api.together.xyz/"
DEFAULT_SUPPLY_ENDPOINT = "https://computer.together.xyz"


class API:
    def __init__(
        self,
        endpoint_url: Optional[str] = None,
        supply_endpoint_url: Optional[str] = None,
        log_level: str = "WARNING",
    ) -> None:
        # Setup logger
        self.logger = get_logger(__name__, log_level=log_level)

        self.together_api_key = os.environ.get("TOGETHER_API_KEY", None)
        if self.together_api_key is None:
            self.logger.critical(
                "TOGETHER_API_KEY not found. Please set it as an environment variable."
            )
            exit_1(self.logger)

        if endpoint_url is None:
            endpoint_url = DEFAULT_ENDPOINT

        if supply_endpoint_url is None:
            supply_endpoint_url = DEFAULT_SUPPLY_ENDPOINT

        self.endpoint_url = endpoint_url
        self.supply_endpoint_url = supply_endpoint_url

    def get_supply(self) -> Dict[str, Any]:
        try:
            response = requests.get(
                self.supply_endpoint_url,
                json={
                    "method": "together_getDepth",
                    "id": 1,
                },
            )
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def get_all_models(self) -> List[str]:
        models = cast(List[str], self.get_supply()["result"].keys())

        models = [str(sub[:-1]) for sub in models]  # remove the ? after the model names

        return models

    def get_available_models(self) -> List[str]:
        res = self.get_supply()
        names = res["result"].keys()
        available_models = [
            name[:-1] for name in names if res["result"][name]["num_asks"] > 0
        ]

        return available_models

    def finetune(self) -> Finetune:
        return Finetune(
            endpoint_url=self.endpoint_url,
        )

    def complete(self, **model_kwargs: Any) -> Inference:
        return Inference(
            endpoint_url=self.endpoint_url,
            **model_kwargs,
        )

    def files(self) -> Files:
        return Files(
            endpoint_url=self.endpoint_url,
        )



import os
import posixpath
import urllib.parse
from typing import Any, Dict, List, Optional, Union

import requests
from tqdm import tqdm

from together.utils import exit_1, get_logger


DEFAULT_ENDPOINT = "https://api.together.xyz/"


class Finetune:
    def __init__(
        self,
        endpoint_url: Optional[str] = None,
        log_level: str = "WARNING",
    ) -> None:
        # Setup logger
        self.logger = get_logger(__name__, log_level=log_level)

        self.together_api_key = os.environ.get("TOGETHER_API_KEY", None)
        if self.together_api_key is None:
            self.logger.critical(
                "TOGETHER_API_KEY not found. Please set it as an environment variable."
            )
            exit_1(self.logger)

        if endpoint_url is None:
            endpoint_url = DEFAULT_ENDPOINT

        self.endpoint_url = urllib.parse.urljoin(endpoint_url, "/v1/fine-tunes/")

    def create_finetune(
        self,
        training_file: str,  # training file_id
        # validation_file: Optional[str] = None,  # validation file_id
        model: Optional[str] = None,
        n_epochs: Optional[int] = 1,
        batch_size: Optional[int] = 32,
        learning_rate: Optional[float] = 0.00001,
        # warmup_steps: Optional[int] = 0,
        # train_warmup_steps: Optional[int] = 0,
        # seq_length: Optional[int] = 2048,
        # seed: Optional[int] = 42,
        # fp16: Optional[bool] = True,
        # checkpoint_steps: Optional[int] = None,
        suffix: Optional[str] = None,
    ) -> Dict[Any, Any]:
        parameter_payload = {
            "training_file": training_file,
            # "validation_file": validation_file,
            "model": model,
            "n_epochs": n_epochs,
            "batch_size": batch_size,
            "learning_rate": learning_rate,
            # "warmup_steps": warmup_steps,
            # "train_warmup_steps": train_warmup_steps,
            # "seq_length": seq_length,
            # "seed": seed,
            # "fp16": fp16,
            "suffix": suffix,
        }

        # HTTP headers for authorization
        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
            "Content-Type": "application/json",
        }

        # send request
        try:
            response = requests.post(
                self.endpoint_url, headers=headers, json=parameter_payload
            )
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def list_finetune(self) -> Dict[Any, Any]:
        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            response = requests.get(self.endpoint_url, headers=headers)
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def retrieve_finetune(self, fine_tune_id: str) -> Dict[Any, Any]:
        retrieve_url = urllib.parse.urljoin(self.endpoint_url, fine_tune_id)

        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            response = requests.get(retrieve_url, headers=headers)
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def cancel_finetune(self, fine_tune_id: str) -> Dict[Any, Any]:
        relative_path = posixpath.join(fine_tune_id, "cancel")
        retrieve_url = urllib.parse.urljoin(self.endpoint_url, relative_path)

        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            response = requests.post(retrieve_url, headers=headers)
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def list_finetune_events(self, fine_tune_id: str) -> Dict[Any, Any]:
        # TODO enable stream
        relative_path = posixpath.join(fine_tune_id, "events")
        retrieve_url = urllib.parse.urljoin(self.endpoint_url, relative_path)

        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        # send request
        try:
            response = requests.get(retrieve_url, headers=headers)
        except requests.exceptions.RequestException as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        try:
            response_json = dict(response.json())
        except Exception as e:
            self.logger.critical(
                f"JSON Error raised: {e}\nResponse status code = {response.status_code}"
            )
            exit_1(self.logger)

        return response_json

    def get_checkpoints(self, fine_tune_id: str) -> List[Dict[str, Any]]:
        try:
            finetune_events = list(
                self.retrieve_finetune(fine_tune_id=fine_tune_id)["events"]
            )
        except Exception as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        saved_events = [i for i in finetune_events if i["type"] in ["CHECKPOINT_SAVE"]]

        return saved_events

    def get_job_status(self, fine_tune_id: str) -> str:
        try:
            job_status = str(
                self.retrieve_finetune(fine_tune_id=fine_tune_id)["status"]
            )
        except Exception as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        return job_status

    def is_final_model_available(self, fine_tune_id: str) -> bool:
        try:
            finetune_events = list(
                self.retrieve_finetune(fine_tune_id=fine_tune_id)["events"]
            )
        except Exception as e:
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        for i in finetune_events:
            if i["type"] in ["JOB_COMPLETE", "JOB_ERROR"]:
                if i["checkpoint_path"] != "":
                    return False
                else:
                    return True
        return False

    def download(
        self,
        fine_tune_id: str,
        output: Union[str, None] = None,
        checkpoint_num: int = -1,
    ) -> str:
        # default to model_output_path name
        if output is None:
            output = (
                self.retrieve_finetune(fine_tune_id)["model_output_path"].split("/")[-1]
                + ".tar.gz"
            )

        model_file_path = urllib.parse.urljoin(
            self.endpoint_url,
            f"/api/finetune/downloadfinetunefile?ft_id={fine_tune_id}",
        )

        self.logger.info(f"Downloading {model_file_path}...")

        headers = {
            "Authorization": f"Bearer {self.together_api_key}",
        }

        try:
            session = requests.Session()

            response = session.get(model_file_path, headers=headers, stream=True)
            response.raise_for_status()

            total_size_in_bytes = int(response.headers.get("content-length", 0))
            block_size = 1024 * 1024  # 1 MB
            progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
            with open(output, "wb") as file:
                for chunk in response.iter_content(block_size):
                    progress_bar.update(len(chunk))
                    file.write(chunk)
            progress_bar.close()
            if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
                self.logger.warning(
                    "Caution: Downloaded file size does not match remote file size."
                )
        except requests.exceptions.RequestException as e:  # This is the correct syntax
            self.logger.critical(f"Response error raised: {e}")
            exit_1(self.logger)

        return output  # this should be null

    # def delete_finetune_model(self, model: str) -> Dict[Any, Any]:
    #     model_url = "https://api.together.xyz/api/models"
    #     delete_url = urllib.parse.urljoin(model_url, model)

    #     headers = {
    #         "Authorization": f"Bearer {self.together_api_key}",
    #     }

    #     # send request
    #     try:
    #         response = requests.delete(delete_url, headers=headers)
    #     except requests.exceptions.RequestException as e:
    #         raise ValueError(f"Error raised by finetune endpoint: {e}")

    #     try:
    #         response_json = dict(response.json())
    #     except Exception as e:
    #         raise ValueError(
    #             f"JSON Error raised. \nResponse status code: {str(response.status_code)}"
    #         )

    #     return response_json



from __future__ import annotations

import argparse
import json
from typing import Any, Dict, List

from together.finetune import Finetune


def extract_time(json_obj: Dict[str, Any]) -> int:
    try:
        return int(json_obj["updated_at"])
    except KeyError:
        return 0


def add_parser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    COMMAND_NAME = "finetune"
    parser = subparsers.add_parser(COMMAND_NAME)

    child_parsers = parser.add_subparsers(required=True)

    _add_create(child_parsers, parents=parents)
    _add_list(child_parsers, parents=parents)
    _add_retrieve(child_parsers, parents=parents)
    _add_list_events(child_parsers, parents=parents)
    _add_cancel(child_parsers, parents=parents)
    _add_download(child_parsers, parents=parents)
    _add_status(child_parsers, parents=parents)
    _add_checkpoints(child_parsers, parents=parents)
    # _add_delete_model(child_parsers)


def _add_create(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    # Create_finetune
    create_finetune_parser = parser.add_parser("create", parents=parents)
    create_finetune_parser.add_argument(
        "--training-file",
        "-t",
        metavar="FILE-ID",
        help="File-ID of an uploaded file that contains training data.",
        required=True,
        type=str,
    )
    # create_finetune_parser.add_argument(
    #     "--validation-file",
    #     "-v",
    #     default=None,
    #     help="The ID of an uploaded file that contains validation data.",
    #     type=str,
    # )
    create_finetune_parser.add_argument(
        "--model",
        "-m",
        metavar="MODEL",
        default=None,
        help="The name of the base model to fine-tune. Default='togethercomputer/RedPajama-INCITE-7B-Chat'.",
        type=str,
    )
    create_finetune_parser.add_argument(
        "--n-epochs",
        "-ne",
        metavar="EPOCHS",
        default=4,
        help="The number of epochs to train the model for. Default=4",
        type=int,
    )
    create_finetune_parser.add_argument(
        "--batch-size",
        "-b",
        metavar="BATCH_SIZE",
        default=32,
        help="The batch size to use for training. Default=32",
        type=int,
    )
    create_finetune_parser.add_argument(
        "--learning-rate",
        "-lr",
        metavar="LEARNING_RATE",
        default=0.00001,
        help="The learning rate multiplier to use for training. Default=0.00001",
        type=float,
    )
    # create_finetune_parser.add_argument(
    #     "--warmup-steps",
    #     "-ws",
    #     default=0,
    #     help="Warmup steps",
    #     type=int,
    # )
    # create_finetune_parser.add_argument(
    #     "--train-warmup-steps",
    #     "-tws",
    #     default=0,
    #     help="Train warmup steps",
    #     type=int,
    # )
    # create_finetune_parser.add_argument(
    #     "--sequence-length",
    #     "-sl",
    #     default=2048,
    #     help="Max sequence length",
    #     type=int,
    # )
    # create_finetune_parser.add_argument(
    #     "--seed",
    #     default=42,
    #     help="Training seed",
    #     type=int,
    # )
    # create_finetune_parser.add_argument(
    #     "--fp32",
    #     help="Enable FP32 training. Defaults to false (FP16 training).",
    #     default=False,
    #     action="store_true",
    # )
    # create_finetune_parser.add_argument(
    #     "--checkpoint-steps",
    #     "-b",
    #     default=0,
    #     help="Number of steps between each checkpoint. Defaults to 0 = checkpoints per epoch.",
    #     type=int,
    # )
    create_finetune_parser.add_argument(
        "--suffix",
        "-s",
        metavar="SUFFIX",
        default=None,
        help="Up to 40 characters that will be added to your fine-tuned model name.",
        type=str,
    )
    create_finetune_parser.set_defaults(func=_run_create)

    # End of create_finetune


def _add_list(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    # List_Finetune
    list_parser = parser.add_parser("list", parents=parents)
    list_parser.set_defaults(func=_run_list)


def _add_retrieve(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    retrieve_finetune_parser = parser.add_parser("retrieve", parents=parents)
    retrieve_finetune_parser.add_argument(
        "fine_tune_id",
        metavar="FINETUNE-ID",
        default=None,
        help="Fine-tuning ID",
        type=str,
    )
    retrieve_finetune_parser.set_defaults(func=_run_retrieve)


def _add_cancel(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    # Cancel Finetune
    cancel_finetune_parser = parser.add_parser("cancel", parents=parents)
    cancel_finetune_parser.add_argument(
        "fine_tune_id",
        metavar="FINETUNE-ID",
        default=None,
        help="Fine-tuning ID",
        type=str,
    )
    cancel_finetune_parser.set_defaults(func=_run_cancel)


def _add_list_events(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    # List finetune events
    list_finetune_events_parser = parser.add_parser("list-events", parents=parents)
    list_finetune_events_parser.add_argument(
        "fine_tune_id",
        metavar="FINETUNE-ID",
        default=None,
        help="Fine-tuning ID",
        type=str,
    )
    list_finetune_events_parser.set_defaults(func=_run_list_events)


def _add_download(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    # List finetune events
    download_parser = parser.add_parser("download", parents=parents)
    download_parser.add_argument(
        "fine_tune_id",
        metavar="FINETUNE-ID",
        default=None,
        help="Fine-tuning ID",
        type=str,
    )
    download_parser.add_argument(
        "--output",
        "-o",
        metavar="FILENAME",
        default=None,
        help="Output filename. Defaults to remote name.",
        type=str,
        required=False,
    )
    # download_parser.add_argument(
    #     "--checkpoint-num",
    #     "-n",
    #     default=-1,
    #     help="Checkpoint number. Defaults to the latest checkpoint = -1.",
    #     type=int,
    #     required=False,
    # )
    download_parser.set_defaults(func=_run_download)


def _add_status(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    # List finetune events
    status_parser = parser.add_parser("status", parents=parents)
    status_parser.add_argument(
        "fine_tune_id",
        metavar="FINETUNE-ID",
        default=None,
        help="Fine-tuning ID",
        type=str,
    )
    status_parser.set_defaults(func=_run_status)


def _add_checkpoints(
    parser: argparse._SubParsersAction[argparse.ArgumentParser],
    parents: List[argparse.ArgumentParser],
) -> None:
    # List finetune events
    checkpoint_parser = parser.add_parser("checkpoints", parents=parents)
    checkpoint_parser.add_argument(
        "fine_tune_id",
        metavar="FINETUNE-ID",
        default=None,
        help="Fine-tuning ID",
        type=str,
    )
    checkpoint_parser.set_defaults(func=_run_checkpoint)


# def _add_delete_model(
#     parser: argparse._SubParsersAction[argparse.ArgumentParser],
# ) -> None:
#     # Delete finetune model
#     delete_finetune_model_parser = parser.add_parser("delete-model")
#     delete_finetune_model_parser.add_argument(
#         "--model",
#         "-m",
#         default=None,
#         help="Model name",
#         type=str,
#         required=True,
#     )
#     delete_finetune_model_parser.set_defaults(func=_run_delete_model)


def _run_create(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)

    response = finetune.create_finetune(
        training_file=args.training_file,  # training file_id
        # validation_file=args.validation_file,  # validation file_id
        model=args.model,
        n_epochs=args.n_epochs,
        batch_size=args.batch_size,
        learning_rate=args.learning_rate,
        # warmup_steps=args.warmup_steps,
        # train_warmup_steps=args.train_warmup_steps,
        # seq_length=args.sequence_length,
        # seed=args.seed,
        # fp16=not args.fp32,
        # checkpoint_steps=args.checkpoint_steps,
        suffix=args.suffix,
    )

    print(json.dumps(response, indent=4))


def _run_list(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)
    response = finetune.list_finetune()
    for item in response["data"]:
        item.pop("events", None)
    response["data"].sort(key=extract_time)
    print(json.dumps(response, indent=4))


def _run_retrieve(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)
    response = finetune.retrieve_finetune(args.fine_tune_id)
    print(json.dumps(response, indent=4))


def _run_cancel(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)
    response = finetune.cancel_finetune(args.fine_tune_id)
    print(json.dumps(response, indent=4))


def _run_list_events(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)
    response = finetune.list_finetune_events(args.fine_tune_id)
    print(json.dumps(response, indent=4))


def _run_download(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)
    args.checkpoint_num = -1  # hardcoded until enabled in remote
    response = finetune.download(args.fine_tune_id, args.output, args.checkpoint_num)
    print(response)


def _run_status(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)
    response = finetune.get_job_status(args.fine_tune_id)
    print(response)


def _run_checkpoint(args: argparse.Namespace) -> None:
    finetune = Finetune(args.endpoint, log_level=args.log)
    checkpoints = finetune.get_checkpoints(args.fine_tune_id)
    print(json.dumps(checkpoints, indent=4))
    print(f"\n{len(checkpoints)} checkpoints found")


# def _run_delete_model(args: argparse.Namespace) -> None:
#     finetune = Finetune(args.endpoint)
#     response = finetune.delete_finetune_model(args.model)
#     print(json.dumps(response))
